@@ -15,38 +15,33 @@
 #    under the License.
 
 import abc
-import collections
+import functools
 import logging
 import pprint
 
 import alembic
 import alembic.autogenerate
 import alembic.migration
-import pkg_resources as pkg
-import six
 import sqlalchemy
-from sqlalchemy.engine import reflection
 import sqlalchemy.exc
-from sqlalchemy import schema
 import sqlalchemy.sql.expression as expr
 import sqlalchemy.types as types
 
-from oslo_db._i18n import _LE
 from oslo_db import exception as exc
+from oslo_db.sqlalchemy import provision
 from oslo_db.sqlalchemy import utils
 
 LOG = logging.getLogger(__name__)
 
 
-@six.add_metaclass(abc.ABCMeta)
-class WalkVersionsMixin(object):
+class WalkVersionsMixin(object, metaclass=abc.ABCMeta):
     """Test mixin to check upgrade and downgrade ability of migration.
 
     This is only suitable for testing of migrate_ migration scripts. An
     abstract class mixin. `INIT_VERSION`, `REPOSITORY` and `migration_api`
     attributes must be implemented in subclasses.
 
-    .. _auxiliary-dynamic-methods: Auxiliary Methods
+    .. _auxiliary-dynamic-methods:
 
     Auxiliary Methods:
 
@@ -125,30 +120,6 @@ class WalkVersionsMixin(object):
         """
         pass
 
-    def _walk_versions(self, snake_walk=False, downgrade=True):
-        """Check if migration upgrades and downgrades successfully.
-
-        DEPRECATED: this function is deprecated and will be removed from
-        oslo.db in a few releases. Please use walk_versions() method instead.
-        """
-        self.walk_versions(snake_walk, downgrade)
-
-    def _migrate_down(self, version, with_data=False):
-        """Migrate down to a previous version of the db.
-
-        DEPRECATED: this function is deprecated and will be removed from
-        oslo.db in a few releases. Please use migrate_down() method instead.
-        """
-        return self.migrate_down(version, with_data)
-
-    def _migrate_up(self, version, with_data=False):
-        """Migrate up to a new version of the db.
-
-        DEPRECATED: this function is deprecated and will be removed from
-        oslo.db in a few releases. Please use migrate_up() method instead.
-        """
-        self.migrate_up(version, with_data)
-
     def walk_versions(self, snake_walk=False, downgrade=True):
         """Check if migration upgrades and downgrades successfully.
 
@@ -167,7 +138,7 @@ class WalkVersionsMixin(object):
 
             If we have ordered migrations 123abc, 456def, 789ghi and we run
             upgrading with the `snake_walk` argument set to `True`, the
-            migrations will be applied in the following order:
+            migrations will be applied in the following order::
 
                 `123abc => 456def => 123abc =>
                  456def => 789ghi => 456def => 789ghi`
@@ -186,7 +157,8 @@ class WalkVersionsMixin(object):
                                                        self.REPOSITORY))
 
         LOG.debug('latest version is %s', self.REPOSITORY.latest)
-        versions = range(self.INIT_VERSION + 1, self.REPOSITORY.latest + 1)
+        versions = range(int(self.INIT_VERSION) + 1,
+                         int(self.REPOSITORY.latest) + 1)
 
         for version in versions:
             # upgrade -> downgrade -> upgrade
@@ -214,7 +186,7 @@ class WalkVersionsMixin(object):
         :type version: str
         :keyword with_data: Whether to verify the absence of changes from
             migration(s) being downgraded, see
-            :ref:`auxiliary-dynamic-methods <Auxiliary Methods>`.
+            :ref:`Auxiliary Methods <auxiliary-dynamic-methods>`.
         :type with_data: Bool
         """
 
@@ -246,7 +218,7 @@ class WalkVersionsMixin(object):
         :param version: id of revision to upgrade.
         :type version: str
         :keyword with_data: Whether to verify the applied changes with data,
-            see :ref:`auxiliary-dynamic-methods <Auxiliary Methods>`.
+            see :ref:`Auxiliary Methods <auxiliary-dynamic-methods>`.
         :type with_data: Bool
         """
         # NOTE(sdague): try block is here because it's impossible to debug
@@ -268,14 +240,13 @@ class WalkVersionsMixin(object):
                 check = getattr(self, "_check_%03d" % version, None)
                 if check:
                     check(self.migrate_engine, data)
-        except exc.DbMigrationError:
-            msg = _LE("Failed to migrate to version %(ver)s on engine %(eng)s")
+        except exc.DBMigrationError:
+            msg = "Failed to migrate to version %(ver)s on engine %(eng)s"
             LOG.error(msg, {"ver": version, "eng": self.migrate_engine})
             raise
 
 
-@six.add_metaclass(abc.ABCMeta)
-class ModelsMigrationsSync(object):
+class ModelsMigrationsSync(object, metaclass=abc.ABCMeta):
     """A helper class for comparison of DB migration scripts and models.
 
     It's intended to be inherited by test cases in target projects. They have
@@ -435,7 +406,19 @@ class ModelsMigrationsSync(object):
         if issubclass(type(meta_type), BOOLEAN_METADATA):
             return not issubclass(type(insp_type), BOOLEAN_SQL)
 
-        return None  # tells alembic to use the default comparison method
+        # Alembic <=0.8.4 do not contain logic of comparing Variant type with
+        # others.
+        if isinstance(meta_type, types.Variant):
+            orig_type = meta_col.type
+            impl_type = meta_type.load_dialect_impl(ctxt.dialect)
+            meta_col.type = impl_type
+            try:
+                return self.compare_type(ctxt, insp_col, meta_col, insp_type,
+                                         impl_type)
+            finally:
+                meta_col.type = orig_type
+
+        return ctxt.impl.compare_type(insp_col, meta_col)
 
     def compare_server_default(self, ctxt, ins_col, meta_col,
                                insp_def, meta_def, rendered_meta_def):
@@ -464,135 +447,29 @@ class ModelsMigrationsSync(object):
         if isinstance(meta_col.type, sqlalchemy.Boolean):
             if meta_def is None or insp_def is None:
                 return meta_def != insp_def
+            insp_def = insp_def.strip("'")
             return not (
-                isinstance(meta_def.arg, expr.True_) and insp_def == "'1'" or
-                isinstance(meta_def.arg, expr.False_) and insp_def == "'0'"
+                isinstance(meta_def.arg, expr.True_) and insp_def == "1" or
+                isinstance(meta_def.arg, expr.False_) and insp_def == "0"
             )
 
-        if isinstance(meta_col.type, sqlalchemy.Integer):
-            if meta_def is None or insp_def is None:
-                return meta_def != insp_def
-            return meta_def.arg != insp_def.split("'")[1]
+    def filter_metadata_diff(self, diff):
+        """Filter changes before assert in test_models_sync().
 
-    @_compare_server_default.dispatch_for('postgresql')
-    def _compare_server_default(bind, meta_col, insp_def, meta_def):
-        if isinstance(meta_col.type, sqlalchemy.Enum):
-            if meta_def is None or insp_def is None:
-                return meta_def != insp_def
-            return insp_def != "'%s'::%s" % (meta_def.arg, meta_col.type.name)
-        elif isinstance(meta_col.type, sqlalchemy.String):
-            if meta_def is None or insp_def is None:
-                return meta_def != insp_def
-            return insp_def != "'%s'::character varying" % meta_def.arg
-
-    def _cleanup(self):
-        engine = self.get_engine()
-        with engine.begin() as conn:
-            inspector = reflection.Inspector.from_engine(engine)
-            metadata = schema.MetaData()
-            tbs = []
-            all_fks = []
-
-            for table_name in inspector.get_table_names():
-                fks = []
-                for fk in inspector.get_foreign_keys(table_name):
-                    if not fk['name']:
-                        continue
-                    fks.append(
-                        schema.ForeignKeyConstraint((), (), name=fk['name'])
-                        )
-                table = schema.Table(table_name, metadata, *fks)
-                tbs.append(table)
-                all_fks.extend(fks)
-
-            for fkc in all_fks:
-                conn.execute(schema.DropConstraint(fkc))
-
-            for table in tbs:
-                conn.execute(schema.DropTable(table))
-
-    FKInfo = collections.namedtuple('fk_info', ['constrained_columns',
-                                                'referred_table',
-                                                'referred_columns'])
-
-    def check_foreign_keys(self, metadata, bind):
-        """Compare foreign keys between model and db table.
-
-        :returns: a list that contains information about:
-
-         * should be a new key added or removed existing,
-         * name of that key,
-         * source table,
-         * referred table,
-         * constrained columns,
-         * referred columns
-
-         Output::
-
-             [('drop_key',
-               'testtbl_fk_check_fkey',
-               'testtbl',
-               fk_info(constrained_columns=(u'fk_check',),
-                       referred_table=u'table',
-                       referred_columns=(u'fk_check',)))]
-
-        DEPRECATED: this function is deprecated and will be removed from
-        oslo.db in a few releases. Alembic autogenerate.compare_metadata()
-        now includes foreign key comparison directly.
+        Allow subclasses to whitelist/blacklist changes. By default, no
+        filtering is performed, changes are returned as is.
 
+        :param diff: a list of differences (see `compare_metadata()` docs for
+            details on format)
+        :returns: a list of differences
         """
-
-        diff = []
-        insp = sqlalchemy.engine.reflection.Inspector.from_engine(bind)
-        # Get all tables from db
-        db_tables = insp.get_table_names()
-        # Get all tables from models
-        model_tables = metadata.tables
-        for table in db_tables:
-            if table not in model_tables:
-                continue
-            # Get all necessary information about key of current table from db
-            fk_db = dict((self._get_fk_info_from_db(i), i['name'])
-                         for i in insp.get_foreign_keys(table))
-            fk_db_set = set(fk_db.keys())
-            # Get all necessary information about key of current table from
-            # models
-            fk_models = dict((self._get_fk_info_from_model(fk), fk)
-                             for fk in model_tables[table].foreign_keys)
-            fk_models_set = set(fk_models.keys())
-            for key in (fk_db_set - fk_models_set):
-                diff.append(('drop_key', fk_db[key], table, key))
-                LOG.info(("Detected removed foreign key %(fk)r on "
-                          "table %(table)r"), {'fk': fk_db[key],
-                                               'table': table})
-            for key in (fk_models_set - fk_db_set):
-                diff.append(('add_key', fk_models[key], table, key))
-                LOG.info((
-                    "Detected added foreign key for column %(fk)r on table "
-                    "%(table)r"), {'fk': fk_models[key].column.name,
-                                   'table': table})
         return diff
 
-    def _get_fk_info_from_db(self, fk):
-        return self.FKInfo(tuple(fk['constrained_columns']),
-                           fk['referred_table'],
-                           tuple(fk['referred_columns']))
-
-    def _get_fk_info_from_model(self, fk):
-        return self.FKInfo((fk.parent.name,), fk.column.table.name,
-                           (fk.column.name,))
-
     def test_models_sync(self):
-        # recent versions of sqlalchemy and alembic are needed for running of
-        # this test, but we already have them in requirements
-        try:
-            pkg.require('sqlalchemy>=0.8.4', 'alembic>=0.6.2')
-        except (pkg.VersionConflict, pkg.DistributionNotFound) as e:
-            self.skipTest('sqlalchemy>=0.8.4 and alembic>=0.6.3 are required'
-                          ' for running of this test: %s' % e)
-
-        # drop all tables after a test run
-        self.addCleanup(self._cleanup)
+        # drop all objects after a test run
+        engine = self.get_engine()
+        backend = provision.Backend(engine.name, engine.url)
+        self.addCleanup(functools.partial(backend.drop_all_objects, engine))
 
         # run migration scripts
         self.db_sync(self.get_engine())
@@ -606,8 +483,8 @@ class ModelsMigrationsSync(object):
             mc = alembic.migration.MigrationContext.configure(conn, opts=opts)
 
             # compare schemas and fail with diff, if it's not empty
-            diff = alembic.autogenerate.compare_metadata(mc,
-                                                         self.get_metadata())
+            diff = self.filter_metadata_diff(
+                alembic.autogenerate.compare_metadata(mc, self.get_metadata()))
             if diff:
                 msg = pprint.pformat(diff, indent=2, width=20)
                 self.fail(
